import socket
import struct
import re

SOURCE_TABLE = {
    0: 'GSI',
    1: 'I8259',
    2: 'IOAPIC',
    3: "HW_EDU"
}

SOURCE_BUCKET_SIZE = 0x10
INT_BUCKET_SIZE = 256


IP = '192.168.135.101'  # Server IP address
PORT = 4444      # Server port

sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
sock.connect((IP, PORT))

endian_test_data = sock.recv(4)

if endian_test_data == b'\x00\x00\x00\x01':
    ENDIAN = 'big'
elif endian_test_data == b'\x01\x00\x00\x00':
    ENDIAN = 'little'

buf = b''
current_pdu = {
    "op": None,
    "length": None,
    "data": None
}

def send_request(op, data=b''):
    len_data = len(data)
    if ENDIAN == 'big':
        packet = struct.pack('>BI', op, len_data) + data
    else:
        packet = struct.pack('<BI', op, len_data) + data
    sock.sendall(packet)
    buf = sock.recv(1024)
    if ENDIAN == 'big':
        current_pdu = {
            "op": buf[0],
            "length": struct.unpack('>I', buf[1:5])[0],
            "data": buf[5:]
        }
    else:
        current_pdu = {
            "op": buf[0],
            "length": struct.unpack('<I', buf[1:5])[0],
            "data": buf[5:]
        }
    while current_pdu["length"] > len(current_pdu["data"]):
        buf = sock.recv(current_pdu["length"] - len(current_pdu["data"]))
        current_pdu["data"] += buf
    pdu = current_pdu
    return pdu

def command_prompt():
    inp = input("(intsvc)> ")
    return inp

def help():
    print("ver - Get version")
    print("ping - Ping")
    print("status - Get status")
    print("stats - Get statistics")
    print("set-int-timer <value> - Set interrupt timer")
    print("set-int <source> <intno> <value> - Set interrupt")
    print("ls - List interrupts")
    print("send <source> <intno> - Send interrupt")
    print("exit - Exit")

def pretty_print_switches(data):
    sw_intercept_timer_int = data[0]
    int_svc_sw_bucket = data[1:]
    print("Intercept timer INT: " + (sw_intercept_timer_int >= 1 and "on" or "off"))
    # We only implemented GSI and HW_EDU
    # Find ranges that equals 0
    intercepted_range = []
    range_start = 0
    range_end = 0
    for i in range(0, INT_BUCKET_SIZE):
        if int_svc_sw_bucket[i] == 0:
            range_end = i
        else:
            if range_end - range_start > 0:
                intercepted_range.append((range_start, range_end))
            range_start = i + 1
    print("GSI(passthrough):")
    if len(intercepted_range) == 0:
        print("  None")
    else:
        for i in intercepted_range:
            print(f"  {i[0]}-{i[1]}")
    print("HW_EDU(passthrough):")
    if (int_svc_sw_bucket[INT_BUCKET_SIZE * 3] == 1):
        print("  None")
    else:
        print("  All")

def pretty_print_interrupts(data):
    print("Interrupts:")
    # typedef struct int_svc_interception_response {
    #     int source;
    #     int intno;
    #     int count;
    # } IntSvcInterceptionResponse;
    # 12 bytes per entry
    if data == b'':
        print("  None")
        return
    for i in range(0, len(data), 12):
        if ENDIAN == 'big':
            source = struct.unpack('>I', data[i:i+4])[0]
            intno = struct.unpack('>I', data[i+4:i+8])[0]
            count = struct.unpack('>I', data[i+8:i+12])[0]
        else:
            source = struct.unpack('<I', data[i:i+4])[0]
            intno = struct.unpack('<I', data[i+4:i+8])[0]
            count = struct.unpack('<I', data[i+8:i+12])[0]
        print(f"  {SOURCE_TABLE[source]}({source}):{intno} - {count}")


def print_response(response):
    op = response["op"]
    length = response["length"]
    data = response["data"]
    if op == 0x00:
        print(f"Version: {int.from_bytes(data, byteorder='big')}")
    elif op == 0x01:
        print(data.decode('utf-8'))
    elif op == 0x02:
        pretty_print_switches(data)
    elif op == 0x03:
        if ENDIAN == 'big':
            print(f"Interrupts on Chip: {struct.unpack('>I', data[0:4])[0]}")
            print(f"Interrupts Inside CPU: {struct.unpack('>I', data[4:8])[0]}")
        else:
            print(f"Interrupts on Chip: {struct.unpack('<I', data[0:4])[0]}")
            print(f"Interrupts Inside CPU: {struct.unpack('<I', data[4:8])[0]}")
    elif op == 0x04:
        print(f"Interrupt timer interception switched " + ("on" if data[0] >= 1 else "off") + ".")
    elif op == 0x05:
        print(f"Interrupt {SOURCE_TABLE[data[0]]}:{data[1]} switched " + ("on" if data[2] >= 1 else "off") + ".")
    elif op == 0x06:
        pretty_print_interrupts(data)
    elif op == 0x07:
        print(data.decode('utf-8'))
            

def main():
    while True:
        try:
            response = b""
            inp = command_prompt()
            inp = re.sub(' +', ' ', inp)
            inp = inp.split(" ")
            cmd = inp[0]
            if cmd == 'ver':
                response = send_request(0x00)
            elif cmd == 'ping':
                response = send_request(0x01)
            elif cmd == 'status':
                response = send_request(0x02)
            elif cmd == 'stats':
                response = send_request(0x03)
            elif cmd == 'set-int-timer':
                data = bytes([int(inp[1])])
                response = send_request(0x04, data)
            elif cmd == 'set-int':
                source = int(inp[1])
                intno = int(inp[2])
                value = int(inp[3])
                data = bytes([source, intno, value])
                response = send_request(0x05, data)
            elif cmd == 'ls':
                response = send_request(0x06)
            elif cmd == 'send':
                id_val = int(inp[1])
                intno = int(inp[2])
                data = bytes([id_val, intno])
                response = send_request(0x07, data)
            elif cmd == 'exit':
                print("Exiting...")
                break
            elif cmd == 'help':
                help()
            else:
                print("Invalid input, type 'help' for help")
            if len(response) > 0:
                print_response(response)
        except IndexError:
            print("Invalid input, type 'help' for help")
        except KeyboardInterrupt:
            print("Exiting...")
            break
        except ValueError:
            print("Invalid input, type 'help' for help")
if __name__ == '__main__':
    main()
