#include "intsvc/intsvc.h"

// NOTE: Works only on linux since we use epoll() for event handling
// Pls switch to select() when porting to other platforms

// Service switch, initialized by qemu_init
static uint8_t int_svc_enabled = 0;

// Timer interrupt interception switch, controlled by the service
static uint8_t int_svc_sw_intercept_timer_int = 0;

// Interrupt switch bucket by source
static uint8_t int_svc_sw_bucket[SOURCE_BUCKET_SIZE][INT_BUCKET_SIZE];

// May impacts performance
static uint8_t int_svc_stat_irqs = 0;

static struct int_svc_options int_svc_options;

static QemuThread int_svc_thread;

// stats
static int int_svc_logged_ints = 0;
static int int_svc_logged_ints_cpu = 0;
// stats lock
static QemuMutex int_svc_ints_stats_lock;     // for other threads
static QemuMutex int_svc_ints_stats_cpu_lock; // for CPU thread, will be critical when we have multiple vCPUs

// bucket
static IntSvcInterceptions int_svc_int_bucket[INT_BUCKET_SIZE];
// queue lock
static QemuMutex int_svc_int_bucket_lock;

// Socket server fd
static int int_svc_server_fd;

// epoll
static int int_svc_epoll_fd;
static struct epoll_event *int_svc_epoll_events;

int int_svc_server_create(void);
void socket_set_nonblocking(int fd);
void client_toggle_timer_int_interception(int fd, uint8_t value);
char *int_svc_package(uint8_t op, uint32_t len, const char *value);
void client_send_version(int fd);
void client_cmd_ping(int fd);
void client_report_interception_status(int fd);
void client_report_interception_stats(int fd);
void client_toggle_other_int_interception_by_source(int fd, uint8_t source, uint8_t intno, uint8_t value);
void client_get_interception_list(int fd);
void client_pass_interception(int fd, int id, int intno);
void handle_client_cmd(IntSvcClientContext *ctx);
void client_process_tlv(int fd, IntSvcProtocolTlv *tlv);
void client_destroy(int fd, IntSvcClientContext *ctx);

void int_svc_parse_options(const char *options)
{
    int_svc_enabled = 1;
    int_svc_stat_irqs = 1;
    // options store
    char *opt = g_strdup(options);
    char **addr = g_strsplit(opt, ":", 2);
    int_svc_options.addr = addr[0];
    int_svc_options.port = atoi(addr[1]);
    // initialize all other variables
    memset(int_svc_sw_bucket, 1, sizeof(uint8_t) * SOURCE_BUCKET_SIZE * INT_BUCKET_SIZE);
    for (int i = 0; i < SOURCE_HW_EDU; i++)
    {
        int_svc_int_bucket[i].requests = g_new0(IntSvcInterruptRequests, INT_BUCKET_SIZE);
        int_svc_int_bucket[i].is_router = 1;
    }
    qemu_mutex_init(&int_svc_int_bucket_lock);
    qemu_mutex_init(&int_svc_ints_stats_lock);
    qemu_mutex_init(&int_svc_ints_stats_cpu_lock);
    if (int_svc_server_create() < 0)
    {
        printf("int_svc: server create failed\n");
        exit(1);
    }
    qemu_thread_create(&int_svc_thread, "int_svc", int_svc_service_run, NULL, QEMU_THREAD_JOINABLE);
}

void socket_set_nonblocking(int fd)
{
    int flags = fcntl(fd, F_GETFL, 0);
    fcntl(fd, F_SETFL, flags | O_NONBLOCK);
}

int int_svc_server_create(void)
{
    int_svc_server_fd = socket(AF_INET, SOCK_STREAM, 0);

    if (int_svc_server_fd < 0)
    {
        printf("int_svc: socket create failed\n");
        return -1;
    }

    // set non-blocking
    socket_set_nonblocking(int_svc_server_fd);

    // set address
    struct sockaddr_in server_addr;
    memset(&server_addr, 0, sizeof(server_addr));
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(int_svc_options.port);
    server_addr.sin_addr.s_addr = inet_addr(int_svc_options.addr);

    // bind & listen
    if (bind(int_svc_server_fd, (struct sockaddr *)&server_addr, sizeof(server_addr)) < 0)
    {
        printf("int_svc: socket bind failed\n");
        return -1;
    }

    if (listen(int_svc_server_fd, 5) < 0)
    {
        printf("int_svc: socket listen failed\n");
        return -1;
    }

    // epoll
    struct epoll_event ev;

    int_svc_epoll_events = g_new0(struct epoll_event, EPOLL_MAX_EVENTS);

    // ctx instead of fd
    IntSvcClientContext *ctx = g_new0(IntSvcClientContext, 1);
    memset(ctx, 0, sizeof(IntSvcClientContext));
    ctx->fd = int_svc_server_fd;

    ev.events = EPOLLIN;
    ev.data.ptr = ctx;

    int_svc_epoll_fd = epoll_create(1);
    if (int_svc_epoll_fd < 0)
    {
        printf("int_svc: epoll create failed\n");
        return -1;
    }

    if (epoll_ctl(int_svc_epoll_fd, EPOLL_CTL_ADD, int_svc_server_fd, &ev) < 0)
    {
        printf("int_svc: epoll ctl failed\n");
        return -1;
    }

    return 0;
}

char *int_svc_package(uint8_t op, uint32_t len, const char *value)
{
    char *buf = g_new0(char, 5 + len);
    buf[0] = op;
    memcpy(buf + 1, &len, 4);
    memcpy(buf + 5, value, len);
    return buf;
}

void client_send_version(int fd)
{
    char *data = g_new0(char, 1);
    data[0] = INT_SVC_VERSION;
    char *buf = int_svc_package(0x00, 1, data);
    write(fd, buf, 6);
    g_free(buf);
    g_free(data);
}

void client_cmd_ping(int fd)
{
    char *buf = int_svc_package(0x01, 4, "pong");
    write(fd, buf, 5 + 4);
    g_free(buf);
}

void client_report_interception_status(int fd)
{
    int len = 0;
    char *data = g_new0(char, sizeof(int_svc_sw_intercept_timer_int) + sizeof(int_svc_sw_bucket));
    data[0] = int_svc_sw_intercept_timer_int;
    len += sizeof(int_svc_sw_intercept_timer_int);
    memcpy(data + len, int_svc_sw_bucket, sizeof(int_svc_sw_bucket));
    len += sizeof(int_svc_sw_bucket);
    char *buf = int_svc_package(0x02, len, data);
    write(fd, buf, 5 + len);
    g_free(buf);
    g_free(data);
}

void client_report_interception_stats(int fd)
{
    int len = 0;
    char *data = g_new0(char, sizeof(int_svc_logged_ints) + sizeof(int_svc_logged_ints_cpu));
    memcpy(data, &int_svc_logged_ints, sizeof(int_svc_logged_ints));
    len += sizeof(int_svc_logged_ints);
    memcpy(data + len, &int_svc_logged_ints_cpu, sizeof(int_svc_logged_ints_cpu));
    len += sizeof(int_svc_logged_ints_cpu);
    char *buf = int_svc_package(0x03, len, data);
    write(fd, buf, 5 + len);
    g_free(buf);
    g_free(data);
}

void client_toggle_timer_int_interception(int fd, uint8_t value)
{
    int_svc_sw_intercept_timer_int = value;
    char *buf = int_svc_package(0x04, 1, (char *)&value);
    write(fd, buf, 5 + 1);
}

void client_toggle_other_int_interception_by_source(int fd, uint8_t source, uint8_t intno, uint8_t value)
{
    if (source >= SOURCE_HW_EDU)
    {
        intno = 0;
    }
    int_svc_sw_bucket[source][intno] = value;
    char *data = g_new0(char, 3);
    data[0] = source;
    data[1] = intno;
    data[2] = value;
    char *buf = int_svc_package(0x05, 3, data);
    write(fd, buf, 5 + 3);
    g_free(buf);
    g_free(data);
}

void client_get_interception_list(int fd)
{
    int len = 0, irqs = 0;
    qemu_mutex_lock(&int_svc_int_bucket_lock);
    for (int i = 0; i < SOURCE_BUCKET_SIZE; i++)
    {
        if (int_svc_int_bucket[i].is_router)
        {
            for (int j = 0; j < INT_BUCKET_SIZE; j++)
            {
                if (int_svc_int_bucket[i].requests[j].is_set)
                {
                    irqs++;
                }
            }
        }
        else
        {
            if (int_svc_int_bucket[i].is_set)
            {
                irqs++;
            }
        }
    }
    // now, we know the length of the data
    char *data = g_new0(char, irqs * sizeof(IntSvcInterceptionResponse));
    IntSvcInterceptionResponse *responses = (IntSvcInterceptionResponse *)data;
    for (int i = 0; i < SOURCE_BUCKET_SIZE; i++)
    {
        if (int_svc_int_bucket[i].is_router)
        {
            for (int j = 0; j < INT_BUCKET_SIZE; j++)
            {
                if (int_svc_int_bucket[i].requests[j].is_set)
                {
                    responses->source = i;
                    responses->intno = j;
                    responses->count = int_svc_int_bucket[i].requests[j].count;
                    responses++;
                    len += sizeof(IntSvcInterceptionResponse);
                }
            }
        }
        else
        {
            if (int_svc_int_bucket[i].is_set)
            {
                responses->source = i;
                responses->intno = 0;
                responses->count = int_svc_int_bucket[i].count;
                responses++;
                len += sizeof(IntSvcInterceptionResponse);
            }
        }
    }
    qemu_mutex_unlock(&int_svc_int_bucket_lock);
    char *buf = int_svc_package(0x06, len, data);
    write(fd, buf, 5 + len);
    g_free(buf);
    g_free(data);
}

void client_pass_interception(int fd, int id, int intno)
{
    if (id >= SOURCE_BUCKET_SIZE)
    {
        char *buf = int_svc_package(0x07, 12, "Out of range");
        write(fd, buf, 5 + 12);
        g_free(buf);
        return;
    }
    if (intno >= INT_BUCKET_SIZE)
    {
        char *buf = int_svc_package(0x07, 12, "Out of range");
        write(fd, buf, 5 + 12);
        g_free(buf);
        return;
    }
    qemu_mutex_lock(&int_svc_int_bucket_lock);
    // Check whether the interception is still valid
    if (int_svc_int_bucket[id].is_router)
    {
        if (!int_svc_int_bucket[id].requests[intno].is_set)
        {
            qemu_mutex_unlock(&int_svc_int_bucket_lock);
            char *buf = int_svc_package(0x07, 20, "No such interception");
            write(fd, buf, 5 + 20);
            g_free(buf);
            return;
        }
    }
    else
    {
        if (!int_svc_int_bucket[id].is_set)
        {
            qemu_mutex_unlock(&int_svc_int_bucket_lock);
            char *buf = int_svc_package(0x07, 20, "No such interception");
            write(fd, buf, 5 + 20);
            g_free(buf);
            return;
        }
    }
    // Pass interception
    switch (id)
    {
    case SOURCE_HW_EDU:
        qemu_mutex_lock_iothread();
        // since it will eventually call gsi_handler, and will lock int_svc_int_bucket_lock, we need to backup data first and unlock
        void *opaque = int_svc_int_bucket[id].opaque;
        qemu_mutex_unlock(&int_svc_int_bucket_lock);
        if (msi_enabled(opaque))
        {
            msi_notify(opaque, 0);
        }
        else
        {
            pci_set_irq(opaque, 1);
        }
        // relock
        qemu_mutex_lock(&int_svc_int_bucket_lock);
        qemu_mutex_unlock_iothread();
        break;
    case SOURCE_GSI:
        gsi_handler_clean(int_svc_int_bucket[id].requests[intno].opaque, intno, 1);
        break;
    default:
        qemu_mutex_unlock(&int_svc_int_bucket_lock);
        char *buf = int_svc_package(0x07, 15, "Not implemented");
        write(fd, buf, 5 + 15);
        g_free(buf);
        return;
    }
    // Clear the interception
    if (int_svc_int_bucket[id].is_router)
    {
        int_svc_int_bucket[id].requests[intno].is_set = 0;
        int_svc_int_bucket[id].requests[intno].count = 0;
    }
    else
    {
        int_svc_int_bucket[id].is_set = 0;
        int_svc_int_bucket[id].count = 0;
    }
    qemu_mutex_unlock(&int_svc_int_bucket_lock);
    char *buf = int_svc_package(0x07, 2, "OK");
    write(fd, buf, 5 + 2);
    g_free(buf);
}

void client_destroy(int fd, IntSvcClientContext *ctx)
{
    shutdown(fd, SHUT_RDWR);
    close(fd);
    g_free(ctx->buf);
    g_free(ctx);
}

void client_process_tlv(int fd, IntSvcProtocolTlv *tlv)
{
    switch (tlv->type)
    {
    case 0x00:
        client_send_version(fd);
        break;
    case 0x01:
        client_cmd_ping(fd);
        break;
    case 0x02:
        client_report_interception_status(fd);
        break;
    case 0x03:
        client_report_interception_stats(fd);
        break;
    case 0x04:
        client_toggle_timer_int_interception(fd, tlv->value[0]);
        break;
    case 0x05:
        client_toggle_other_int_interception_by_source(fd, tlv->value[0], tlv->value[1], tlv->value[2]);
        break;
    case 0x06:
        client_get_interception_list(fd);
        break;
    case 0x07:
        client_pass_interception(fd, tlv->value[0], tlv->value[1]);
        break;
    }
}

void handle_client_cmd(IntSvcClientContext *ctx)
{
    int fd = ctx->fd;
    uint8_t *buf = g_new0(uint8_t, 1024);
    int n = read(fd, buf, 1024);
    if (n < 0)
    {
        client_destroy(fd, ctx);
        goto int_svc_client_recv_exit;
    }
    // client closed
    if (n == 0)
    {
        client_destroy(fd, ctx);
        goto int_svc_client_recv_exit;
    }
    // check if the buffer is full
    while (n + ctx->buf_size > ctx->buf_limit)
    {
        // resize the buffer
        ctx->buf_limit *= 2;
        ctx->buf = g_realloc(ctx->buf, ctx->buf_limit);
    }
    // copy the data
    memcpy(ctx->buf + ctx->buf_size, buf, n);
    ctx->buf_size += n;
    // process tlv
    for (;;)
    {
        if (ctx->buf_size < 5)
        {
            goto int_svc_client_recv_exit;
        }
        // check if the tlv is complete
        if (ctx->is_tlv_set == 0)
        {
            ctx->tlv.type = ctx->buf[0];
            ctx->tlv.length = ctx->buf[1];
            ctx->tlv.value = g_new0(uint8_t, ctx->tlv.length);
            ctx->is_tlv_set = 1;
        }
        if (ctx->buf_size < ctx->tlv.length + 5)
        {
            goto int_svc_client_recv_exit;
        }
        // copy the value
        memcpy(ctx->tlv.value, ctx->buf + 5, ctx->tlv.length);
        // process the tlv
        client_process_tlv(fd, &ctx->tlv);
        // reset the tlv
        ctx->is_tlv_set = 0;
        ctx->buf_size -= ctx->tlv.length + 5;
        memcpy(ctx->buf, ctx->buf + ctx->tlv.length + 5, ctx->buf_size);
    }
int_svc_client_recv_exit:
    g_free(buf);
}

void *int_svc_service_run(void *arg)
{
    int nfds, connfd;
    struct epoll_event ev;
    for (;;)
    {
        if (int_svc_enabled == 0)
        {
            return NULL;
        }
        nfds = epoll_wait(int_svc_epoll_fd, int_svc_epoll_events, EPOLL_MAX_EVENTS, -1);
        if (nfds < 0)
        {
            printf("int_svc: epoll wait failed\n");
            exit(1);
        }

        for (int i = 0; i < nfds; i++)
        {
            IntSvcClientContext *iter_ctx = int_svc_epoll_events[i].data.ptr;
            if (iter_ctx->fd == int_svc_server_fd)
            {
                connfd = accept(int_svc_server_fd, (struct sockaddr *)NULL, NULL);
                if (connfd < 0)
                {
                    printf("int_svc: accept failed\n");
                    exit(1);
                }
                socket_set_nonblocking(connfd);
                // send endian test data
                uint32_t endian_test = 1;
                write(connfd, &endian_test, sizeof(uint32_t));
                ev.events = EPOLLIN | EPOLLET;
                IntSvcClientContext *ctx = g_new0(IntSvcClientContext, 1);
                memset(ctx, 0, sizeof(IntSvcClientContext));
                ctx->buf = g_new0(uint8_t, 1024);
                ctx->buf_limit = 1024;
                ctx->fd = connfd;
                ev.data.ptr = ctx;
                if (epoll_ctl(int_svc_epoll_fd, EPOLL_CTL_ADD, connfd, &ev) < 0)
                {
                    printf("int_svc: epoll ctl failed\n");
                    exit(1);
                }
            }
            else
            {
                handle_client_cmd(iter_ctx);
            }
        }
    }
}

void int_svc_log_int_with_source(int source, int intno, int level)
{
    if (level == 0)
    {
        return;
    }
    if (likely(!int_svc_enabled))
    {
        return;
    }
    if (unlikely(!int_svc_stat_irqs))
    {
        return;
    }
    qemu_mutex_lock(&int_svc_ints_stats_lock);
    int_svc_logged_ints++;
    qemu_mutex_unlock(&int_svc_ints_stats_lock);
}

void int_svc_log_int_cpu(int intvec)
{
    if (likely(!int_svc_enabled))
    {
        return;
    }
    if (unlikely(!int_svc_stat_irqs))
    {
        return;
    }
    qemu_mutex_lock(&int_svc_ints_stats_cpu_lock);
    int_svc_logged_ints_cpu++;
    qemu_mutex_unlock(&int_svc_ints_stats_cpu_lock);
}

int int_svc_timer_int_intercept(void)
{
    if (unlikely(!int_svc_enabled))
    {
        return 0;
    }
    if (unlikely(!int_svc_sw_intercept_timer_int))
    {
        return 0;
    }
    return 1;
}

int int_svc_intercept_other_int(int source, void *opaque, int irq, int level)
{
    // Optimize for the case that the interception is disabled
    if (unlikely(!int_svc_enabled))
    {
        return 0;
    }
    // IRQ Clear, pass
    if (level == 0)
    {
        return 0;
    }

    // Now, let's get back on track
    qemu_mutex_lock(&int_svc_int_bucket_lock);
    // Check whether the interception is enabled
    if (source < SOURCE_HW_EDU)
    {
        if (int_svc_sw_bucket[source][irq] == 0)
        {
            qemu_mutex_unlock(&int_svc_int_bucket_lock);
            return 0;
        }
    }
    else
    {
        if (int_svc_sw_bucket[source][0] == 0)
        {
            qemu_mutex_unlock(&int_svc_int_bucket_lock);
            return 0;
        }
    }
    // Stat and intercept
    if (source < SOURCE_HW_EDU)
    {
        int_svc_int_bucket[source].requests[irq].is_set = 1;
        int_svc_int_bucket[source].requests[irq].count++;
        int_svc_int_bucket[source].requests[irq].opaque = opaque;
    }
    else
    {
        int_svc_int_bucket[source].is_set = 1;
        int_svc_int_bucket[source].count++;
        int_svc_int_bucket[source].opaque = opaque;
    }
    qemu_mutex_unlock(&int_svc_int_bucket_lock);
    return 1;
}

void int_svc_service_clean_up(void)
{
    if (int_svc_enabled)
    {
        int_svc_enabled = 0;
        close(int_svc_server_fd);
        // Just wait for the thread to exit
        // qemu_thread_join(&int_svc_thread);
        qemu_mutex_destroy(&int_svc_int_bucket_lock);
        qemu_mutex_destroy(&int_svc_ints_stats_lock);
        qemu_mutex_destroy(&int_svc_ints_stats_cpu_lock);
        g_free(int_svc_epoll_events);
    }
}